%===================================%
%  Martin Bernasconi                %
%                                   %
%  TU Darmstadt                     %
%                                   %
%  Fuzzy-C-Means V1.4               %
%===================================%
%
% v0 = fuzzyCMeans(m, c, alpha, error)
%
% Der Algorithmus berechnet zu der Punktkoordinatenmatrix
%
% m = [x11 x12 ... x1n
%            ...
%      xd1 xd2 ... xdn]
%
% 2 <= c Clusterpunkte, wobei d die Dimension des Problems ist. 
% Als Kontrastparameter wird 1 <= alpha < Inf bergeben, 
% sowie ein Abbruchfehler error.
% Das Programm gibt dann die Clustermittelpunkte im Vektor
%
% v0 = [x11 x12 ... x1c
%           ...
%       xd1 xd2 ... xdc]
%
% zurck. Bei Bedarf knnen im Quelltext Schalter zur Ausgabe
% von Clusterkoordinaten, Fehler und Zugeherigkeitsmatrix
% gesetzt werden.


function v0 = fuzzyCMeans(m, c, alpha, error)
% fehlerrueckgabewert
v0 = 0;

% Ausgabeswitches
switchCluster = 0;
switchU = 0;
switchError = 0;
switchDebug = 0;

% wenn anzahl Cluster c < 2, dann macht berechnung keinen Sinn
if c < 2,
    disp('========= FEHLER =========');
    disp('Anzahl der Cluster ausser Wertebereich: 2 < c < anzPunkte !');
    disp(' ');
    return;
end

% wenn alpha < 1, fehlermeldung
if alpha < 1,
    disp('========= FEHLER =========');
    disp('alpha-Wert ausser Wertebereich:  1 <= alpha < Inf !');
    disp(' ');
    return;
end


[dim, anzPunkte] = size(m);

% wenn anzPunkte < 2, keine berechnung mglich
if anzPunkte < 2,
    disp('========= FEHLER =========');
    disp('Anzahl der Punkte ausser Wertebereich: 2 <= anzPunkte !');
    disp(' ');
    return;
end

% wenn c > anzPunkte, anpassen
if c > anzPunkte,
    disp('========= FEHLER =========');
    disp('Anzahl der Cluster wurde auf Anzahl der Punkte reduziert!');
    disp(' ');
    c = anzPunkte;
end

% initialisiere U
U1 = rand(c, anzPunkte);
U1 = normiereU(U1, anzPunkte);
U0 = Inf(c, anzPunkte);

% kopfgesteuerte schleife
% index 0 = r, index 1 = r+1
% 1. schritt: abbruchbedingung schon erreicht?
while max(max(abs(U1-U0))) > error,
    
    % 2. schritt: berechnung der clusterzentren
    v0 = [];
    for i = 1:c,
        for d = 1:dim,
            zaehler(d) = m(d,:)*( U1(i,:).^alpha )';
        end
        nenner = sum(U1(i,:).^alpha);
        
        % anfangswerte fuehrten zu keinem akzeptablen ergebnis
        % ganze berechnung nochmal !
        if nenner == 0,
            v0 = fuzzyCMeans(m, c, alpha, error);
            return;
        end
        v0 = [v0 [zaehler/nenner]'];
        
    end %for
   
    if switchDebug == 1, v0, end
    
    % 3. schritt: berechne die c*l abstaende
    d = []; % = [m1v1 m1v2 ...
            %    m2v1 m2v2 ...
            %    ...       mlvc]
    for j = 1:anzPunkte,
        dNeueZeile = [];
        for i = 1:c,
            diff = m(:,j) - v0(:,i);
            dNeueZeile = [dNeueZeile sqrt(diff'*diff)];
        end %for
        d = [d; dNeueZeile];
    end; %for
    
    if switchDebug == 1, d, end
    
    % 4. schritt: berechne U(r+1) = U1
    % zunaechst separieren in M und Q
    MQ = [];
    for j = 1:anzPunkte,
        MQNeueZeile = [];
        for i = 1:c,
            % wenn distanz zu cluster = 0...
            if d(j,i) == 0,
                % merke dir den cluster = schlechte m(j)'s
                MQNeueZeile = [MQNeueZeile i];
            else
                % sonst schreib null rein = gute m(j)'s
                MQNeueZeile = [MQNeueZeile 0];
            end %if
        end %for
        MQ = [MQ; MQNeueZeile];
    end %for
    
    if switchDebug == 1, MQ, end
    
    % U0 wird nun zu U1, damit U1 neu beschrieben werden kann
    U0 = U1;
    
    % alpha = 1 hier nicht verwendbar, daher einfach nahe eins machen
    if alpha == 1,
        exponent = 10000;
    else
        exponent = 2 / (alpha-1);
    end
    
    for j = 1:anzPunkte,
        % Fall 1: M(j) = {}
        % d.h. kein boeser cluster drin, daher ist max = 0
        if max(MQ(j,:)) == 0,
            for i = 1:c,
                summe = 0;
                for k = 1:c,
                    summe = summe + (d(j,i)/d(j,k))^exponent;
                end %for
                U1(i,j) = 1/summe;
            end %for
            
        % Fall 2: M(j) > {}
        else
            for i = 1:c,
                % fuer alle i aus Q(j):
                if MQ(j,i) == 0, 
                    U1(i,j) = 0; 
                % fuer alle i aus M(j):
                else
                    U1(i,j) = 1; % wird spaeter normiert
                end %if
            end %for
        end %if
    end %for
    
    U1 = normiereU(U1, anzPunkte);

    if switchDebug == 1, U0, end
    if switchDebug == 1, U1, end
    
end % while

% ausgaben
if switchCluster == 1, v0, end;
if switchU == 1, U1, end;
if switchError == 1, max(max(abs(U1-U0))), end;

end % function


% ==== normiert eine Matrix U, so dass die Eintrge einer
%      Spalte als Summe 1 ergeben
function nU = normiereU(U, anzPunkte)
for i = 1:anzPunkte,
    summe = sum(U(:,i));
    U(:,i) = U(:,i) / summe;
end
nU = U;
end

